import greenfoot.*;
import java.awt.Color;
import java.util.ArrayList;
import java.util.List;
/**
 * Displays a ray-casted 3d scene.
 * 
 * @author Duta
 * @version 9/3/2013
 */
public class TheWorld extends World {
    private static final double
        AMBIENT = 0.05,
        GAMMA_CORRECTION = 1/2.2;
    private List<WorldObject> objects;
    private Vector light;
    private Vector cameraPosition;
    private double scaleFactor;
    
    public TheWorld() {
        super(500, 500, 1);
        
        // Calculate the scale factor
        scaleFactor = Math.min(getWidth(), getHeight())/10.0;
        
        // Set up the lighting.
        light = new Vector(0, -8, 10);
        
        // Set up the camera.
        cameraPosition = new Vector(0, 0, 10);
        
        // Put some objects into the world
        objects = new ArrayList<WorldObject>();
        generateWorld();
        
        // Make a background.
        GreenfootImage img = new GreenfootImage(
            getWidth(),
            getHeight());
        img.setColor(Color.BLACK);
        img.fill();
        GreenfootImage img2 = new GreenfootImage(
            "Raytracer by Duta\n" +
            "Click to render.",
            50,
            Color.WHITE,
            Color.BLACK);
        img.drawImage(
            img2,
            img.getWidth()/2 - img2.getWidth()/2,
            img.getHeight()/2 - img2.getHeight()/2);
        setBackground(img);
    }
    
    public void act() {
        if(Greenfoot.mousePressed(null)) {
            long calcTime = 0;
            long drawTime = 0;
            
            calcTime = System.currentTimeMillis();
            Color[][] cols = render();
            calcTime = System.currentTimeMillis() - calcTime;
            
            drawTime = System.currentTimeMillis();
            paint(cols);
            drawTime = System.currentTimeMillis() - drawTime;
            
            GreenfootImage img = getBackground();
            GreenfootImage img2 = new GreenfootImage(
                "Calculation time: " + calcTime + "ms\n" +
                "Draw time: " + drawTime + "ms",
                30,
                Color.CYAN,
                new Color(0, 0, 0, 0));
            img.drawImage(
                img2,
                img.getWidth()/2 - img2.getWidth()/2,
                img.getHeight() - img2.getHeight());
        }
    }
    
    private void generateWorld() {
        // Small green ball
        objects.add(new Sphere(
            new Vector(-4, 0, 0),
            2,
            new Color(0, 255, 0)));
        // Medium red ball
        objects.add(new Sphere(
            new Vector(2, 0, 0),
            3.5,
            new Color(255, 0, 0)));
        // Black flat surface
        objects.add(new Plane(
            new Vector(0, 0, -12),
            new Vector(0, -1, 1),
            new Color(255, 255, 255)));
    }
    
    private Color[][] render() {
        int w = getWidth();
        int h = getHeight();
        Color[][] cols = new Color[w][h];
        // For each pixel
        for(int x = 0; x < w; x++)
        for(int y = 0; y < h; y++) {
            //Fire a ray.
            Ray ray = new Ray(
                // From the camera position.
                cameraPosition,
                // Outwards.
                new Vector(
                    x/scaleFactor-5,
                    y/scaleFactor-5,
                    5).subtract(cameraPosition).getNormal());
            // And see what it hits.
            cols[x][y] = trace(ray);
        }
        return cols;
    }
    
    private void paint(Color[][] cols) {
        for(int x = 0; x < getWidth(); x++)
        for(int y = 0; y < getHeight(); y++) {
            getBackground().setColorAt(x, y, cols[x][y]);
        }
    }
    
    private Intersection testRay(Ray ray) {
        return testRay(ray, null);
    }
    
    private Intersection testRay(Ray ray, WorldObject ignore) {
        // At first, assume no intersection
        Intersection inter = new Intersection(
            new Vector(0, 0, 0),
            -1,
            new Vector(0, 0, 0),
            null);
        
        // For each object
        for(WorldObject o : objects) {
            // Unless we're supposed to be ignoring this one
            if(o == ignore) continue;
            // Get its intersection with the ray
            Intersection inter2 = o.getIntersection(ray);
            double d1 = inter.getDistance();
            double d2 = inter2.getDistance();
            // If there's an intersection and
            // either there wasn't one found before
            // or this one is closer, update the
            // closest intersection to be the new one.
            if(d2 > 0 && (d1 < 0 || d1 > d2)) {
                inter = inter2;
            }
        }
        
        return inter;
    }
    
    private Color trace(Ray ray) {
        // Find the closest intersection this ray has
        Intersection inter = testRay(ray);
        // If there was no intersection
        if(inter.getDistance() < 0) {
            // Return the base colour.
            return new Color(
                (int)(AMBIENT*255),
                (int)(AMBIENT*255),
                (int)(AMBIENT*255));
        }
        
        // From this point on we know
        // there was an object hit.
        
        // Get the vector from the light
        // to the point of intersection
        Vector v = light.subtract(inter.getPoint());
        // If our ray hit the surface of an object,
        // but our light isn't shining on it
        if(inter.getNormal().dotProduct(v) < 0) {
            // Use the colour of the object
            // adjusted to be darker. How much
            // darker it is depends on the amount
            // of ambient light there is.
            Color c = inter.getObject().getColor();
            return new Color(
                (int)(c.getRed()*AMBIENT),
                (int)(c.getGreen()*AMBIENT),
                (int)(c.getBlue()*AMBIENT));
        } else {
            // Otherwise, fire a ray checking if
            // any other objects are blocking the
            // light from reaching this point
            // (i.e. check if it is in shadow)
            Ray lightRay = new Ray(
                inter.getPoint(),
                v.getNormal());
            Color c = inter.getObject().getColor();
            // If nothing's blocking it
            // (so, it's not in shadow)
            if(testRay(lightRay, inter.getObject()).getDistance() == -1) {
                // Find the distance from the
                // light to the surface.
                double mag = v.getMagnitude();
                // Use the distance to work
                // out the light intensity.
                double lightIntensity =
                    1000.0/(4*Math.PI*mag*mag);
                double d = Math.max(
                    inter.getNormal().getNormal().dotProduct(
                        v.getNormal().multiply(lightIntensity)),
                    AMBIENT);
                return new Color(
                    Math.min((int)(c.getRed()*d), 255),
                    Math.min((int)(c.getGreen()*d), 255),
                    Math.min((int)(c.getBlue()*d), 255));
            } else {
                // Otherwise, it's in shadow.
                // Use the colour of the object
                // adjusted to be darker. How much
                // darker it is depends on the amount
                // of ambient light there is.
                return new Color(
                    (int)(c.getRed()*AMBIENT),
                    (int)(c.getGreen()*AMBIENT),
                    (int)(c.getBlue()*AMBIENT));
            }
        }
    }
    
    public Color gammaCorrect(Color c) {
        return new Color(
            (int)(255*Math.pow(c.getRed()/255.0, GAMMA_CORRECTION)),
            (int)(255*Math.pow(c.getGreen()/255.0, GAMMA_CORRECTION)),
            (int)(255*Math.pow(c.getBlue()/255.0, GAMMA_CORRECTION)));
    }
}
